"""Functions copied from matflow-defdap package."""
from __future__ import annotations
from pathlib import Path

import defdap.ebsd as ebsd
from defdap.quat import Quat
import numpy as np
from numpy.typing import NDArray
from scipy.stats import mode
from scipy.ndimage import zoom


def load_microstructure_EBSD(
    EBSD: dict,
    root_path: str,
    scaling_factor: float,
) -> dict:
    ebsd_map = _load_EBSD_map(
        root_path=root_path,
        ebsd_filename=EBSD["filename"],
        ebsd_boundary_tol=EBSD.get("boundary_tol", 10),
        ebsd_min_grain_size=EBSD.get("min_grain_size", 10),
    )

    EBSD_image = _get_EBSD_image(ebsd_map, scaling_factor)

    return {"microstructure_image": EBSD_image}


def _load_EBSD_map(
    root_path: str, ebsd_filename: str, ebsd_boundary_tol: float, ebsd_min_grain_size: int
) -> ebsd.Map:
    "Load EBSD map and detect grains."

    ebsd_map = ebsd.Map(Path(root_path).joinpath(ebsd_filename))

    # check for non-indexed points
    if np.count_nonzero(ebsd_map.phaseArray == 0) != 0:
        raise ValueError("EBSD map contains non-indexed points.")

    ebsd_map.buildQuatArray()

    ebsd_map.findBoundaries(boundDef=ebsd_boundary_tol)
    ebsd_map.findGrains(minGrainSize=ebsd_min_grain_size)
    ebsd_map.calcGrainAvOris()

    return ebsd_map


def _get_EBSD_image(ebsd_map: ebsd.Map, scaling_factor: float) -> dict:
    # Construct an array of Euler angles
    grain_quats = np.empty((len(ebsd_map), 4))

    # Transformation orientations from EBSD orientation reference frame
    # to EBSD spatial reference frame
    frame_transform = Quat.fromAxisAngle(np.array((1, 0, 0)), np.pi)

    if ebsd_map.crystalSym == "hexagonal":
        # Convert hex convention from y // a2 of EBSD map to x // a1 for DAMASK
        hex_transform = Quat.fromAxisAngle(np.array([0, 0, 1]), -np.pi / 6)
        for i, grain in enumerate(ebsd_map):
            grain_quats[i] = (hex_transform * grain.refOri * frame_transform).quatCoef

    else:
        for i, grain in enumerate(ebsd_map):
            grain_quats[i] = (grain.refOri * frame_transform).quatCoef

    # Filter out -2 (too small grains) values in the grain image
    grain_image = ebsd_map.grains
    _remove_small_grain_points(grain_image)

    # scale down image if needed
    if scaling_factor != 1:
        grain_image = zoom(
            grain_image, scaling_factor, order=0, prefilter=False, mode="nearest"
        )

    # downstream expects grain numbering to start at 0 not 1
    grain_image -= 1

    # set grain IDs as contiguous range:
    uniq, inv = np.unique(grain_image.reshape(-1, order="C"), return_inverse=True)
    grain_quats = grain_quats[uniq]
    grain_image = np.reshape(inv, grain_image.shape, order="C")
    grain_phases = np.array([grain.phaseID.item() for grain in ebsd_map])[uniq]

    EBSD_image = {
        "orientations": {
            "type": "quat",
            "unit_cell_alignment": {"x": "a"},
            "quaternions": grain_quats,
            "P": 1,  # DefDAP uses P=+1 (e.g see `defdap.quat.Quat.__mul__`)
            "quat_component_ordering": "scalar-vector",
        },
        "grains": grain_image,
        "scale": ebsd_map.scale,
        "phase_labels": [phase.name for phase in ebsd_map.phases],
        "grain_phases": grain_phases,
    }

    return EBSD_image


def _select_area(
    i: int, j: int, grain_image: NDArray, kind: str | None = None
) -> tuple[NDArray, int]:
    on_edge = 0

    if kind == "4_closest":
        area = []
        if i != 0:
            area.append(grain_image[i - 1, j])
            on_edge += 1
        elif i != grain_image.shape[0] - 1:
            area.append(grain_image[i + 1, j])
            on_edge += 1

        if j != 0:
            area.append(grain_image[i, j - 1])
            on_edge += 1
        elif j != grain_image.shape[1] - 1:
            area.append(grain_image[i, j + 1])
            on_edge += 1

        area = np.array(area)

    else:
        i_min, i_max = 1, 1
        j_min, j_max = 1, 1

        if i == 0:
            i_min = 0
            on_edge += 1
        elif i == grain_image.shape[0] - 1:
            i_max = 0
            on_edge += 1

        if j == 0:
            j_min = 0
            on_edge += 1
        elif j == grain_image.shape[1] - 1:
            j_max = 0
            on_edge += 1

        # select 3x3 region around point
        area = grain_image[i - i_min : i + i_max + 1, j - j_min : j + j_max + 1]

    return area, on_edge


def _remove_small_grain_points(grain_image: NDArray, max_iterations: int = 200):
    # num_neighbours - must have at least this many pixels surrounding
    # start checking for 8 neighbours, then 7 until 2
    for num_neighbours in range(8, 0, -1):
        print(
            f"Starting iterations with at least {num_neighbours} equal "
            f"neighbour pixels"
        )

        force = num_neighbours == 1

        num_bad_prev = 0
        iteration = 0
        while True:
            num_bad = np.count_nonzero(grain_image == -2)
            if num_bad == 0:
                # No bad values left, done
                print("All bad points removed.")
                return
            elif num_bad == num_bad_prev:
                # Not removing any more
                print("Number of bad points is not decreasing!")
                break
            if iteration > max_iterations:
                print("Max iterations.")
                break

            iteration += 1
            if force:
                print("All bad points must die! Picking largest neighbour.")
            else:
                print("Starting iteration {}, num bad: {}".format(iteration, num_bad))

            grain_image_new = np.copy(grain_image)
            for i, j in zip(*np.where(grain_image == -2)):
                kind = "4_closest" if force else None
                area, on_edge = _select_area(i, j, grain_image, kind=kind)
                area = area.flatten()
                area = area[area > 0]  # remove -1 and -2

                if force:
                    sizes = [np.count_nonzero(grain_image == k) for k in area]
                    grain_image_new[i, j] = area[np.argsort(sizes)[-1]]

                else:
                    mode_vals, mode_counts = mode(area, keepdims=True)
                    for mode_val, mode_count in zip(mode_vals, mode_counts):
                        if mode_count >= num_neighbours:
                            grain_image_new[i, j] = mode_val
                            break

            num_bad_prev = num_bad
            # [:, :] required to update the array passed in
            grain_image[:, :] = grain_image_new



from __future__ import annotations
from formable.tensile_test import TensileTest
from formable.utils import read_non_uniform_csv


def read_tensile_test_CSV(
    CSV_file_path: str,
    CSV_arguments: dict,
    eng_stress_col_index: int,
    eng_strain_col_index: int,
    true_stress_col_index: int,
    true_strain_col_index: int,
    stress_units: str,
    plastic_range: list[float],
) -> dict:
    _, CSV_data = read_non_uniform_csv(
        CSV_file_path,
        delimiter=CSV_arguments.get("delimiter"),
        skip_rows=CSV_arguments.get("skip_rows"),
        header_row=CSV_arguments.get("header_row"),
    )

    eng_strain, eng_stress = None, None
    true_strain, true_stress = None, None

    if eng_strain_col_index is not None:
        eng_strain = CSV_data[:, eng_strain_col_index]
        eng_stress = CSV_data[:, eng_stress_col_index]
        stress = eng_stress

    elif true_strain_col_index is not None:
        true_strain = CSV_data[:, true_strain_col_index]
        true_stress = CSV_data[:, true_stress_col_index]
        stress = true_stress

    if stress_units.strip().upper() == "MPA":
        stress *= 1e6
    elif stress_units.strip().upper() == "GPA":
        stress *= 1e9

    tensile_test_obj = TensileTest(
        eng_stress=eng_stress,
        eng_strain=eng_strain,
        true_stress=true_stress,
        true_strain=true_strain,
        plastic_range=plastic_range,
    )

    tensile_test = tensile_test_obj.to_dict()

    return {"tensile_test": tensile_test}



"""
Schemes for describing seeds for growing crystals.
"""
from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing_extensions import Self, TypeGuard

import numpy as np
from numpy.typing import NDArray, ArrayLike
from hpcflow.sdk.core.parameters import ParameterValue
from matflow.param_classes.orientations import Orientations


@dataclass
class MicrostructureSeeds(ParameterValue):
    """
    The seeds for crystalline microstructure.
    """

    _typ = "microstructure_seeds"

    #: The positions of the seeds.
    position: NDArray
    #: The size of box containing the seeds.
    box_size: NDArray
    #: Label for the phase.
    phase_label: str
    #: Orientation data.
    orientations: Orientations | None = None
    #: Seed for the random number generator, if used.
    random_seed: int | None = None

    @staticmethod
    def __is_dict(value) -> TypeGuard[dict[str, Any]]:
        # TypeGuard, not TypeIs; gets the correct type semantics
        return isinstance(value, dict)

    def __post_init__(self) -> None:
        self.box_size = np.asarray(self.box_size)
        self.position = np.asarray(self.position)
        if self.__is_dict(self.orientations):
            self.orientations = Orientations(**self.orientations)
        elif not self.orientations:
            self.orientations = Orientations.from_random(number=self.num_seeds)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, self.__class__)
            and self.position.shape == other.position.shape
            and np.allclose(self.position, other.position)
            and self.box_size.shape == other.box_size.shape
            and np.allclose(self.box_size, other.box_size)
            and self.orientations == other.orientations
            and self.phase_label == other.phase_label
            and self.random_seed == other.random_seed
        )

    @classmethod
    def from_JSON_like(
        cls, position: ArrayLike, orientations: dict | None = None, **kwargs
    ) -> Self:
        """For custom initialisation via YAML or JSON."""
        # TODO: is this needed?
        if orientations:
            orient = Orientations.from_JSON_like(**orientations)
        else:
            orient = None
        return cls(position=np.asarray(position), orientations=orient, **kwargs)

    @property
    def num_seeds(self) -> int:
        """
        The number of seeds.
        """
        return self.position.shape[0]

    @staticmethod
    def __normalize_orientations(
        orientations: Orientations | dict[str, Any] | None
    ) -> Orientations | None:
        if orientations is None:
            return None
        if isinstance(orientations, dict):
            return Orientations(**orientations)
        return orientations

    @classmethod
    def from_random(
        cls,
        num_seeds: int,
        box_size: NDArray,
        phase_label: str,
        *,
        random_seed: int | None = None,
        orientations: Orientations | dict[str, Any] | None = None,
    ) -> Self:
        """
        Generate a random microstructure.

        Parameters
        ----------
        num_seeds
            The number of seeds for the microstructure.
        box_size
            The size of box containing the microstructure.
        phase_label
            Label for the microstructure.
        random_seed
            Seed for the random number generator.
        orientations
            Orientation information. If omitted, random.
        """
        # TODO: ensure unique seeds points wrt to grid cells
        box_size = np.asarray(box_size)
        rng = np.random.default_rng(seed=random_seed)
        position = rng.random((num_seeds, box_size.size)) * box_size
        return cls(
            position=position,
            box_size=box_size,
            phase_label=phase_label,
            orientations=cls.__normalize_orientations(orientations),
            random_seed=random_seed,
        )

    @classmethod
    def from_file(
        cls,
        path: str,
        box_size: NDArray,
        phase_label: str,
        *,
        number: int | None = None,
        start_index: int = 0,
        delimiter: str = " ",
    ) -> Self:
        """
        Load a microstructure definition from a text file.

        Parameters
        ----------
        path
            Path to the file to load from.
        box_size
            The size of box containing the microstructure.
        phase_label
            Label for the microstructure.
        number
            Number of seeds to read from the file.
        start_index
            The line number of the file that the seeds start at.
            Allows skipping headers.
        delimiter
            The delimiter separating values in the file.
            Defaults to space, but commas and tabs are also sensible
            (and correspond to CSV and TSV files respectively).
        """
        data: list[list[float]] = []
        with Path(path).open("rt") as fh:
            for idx, line in enumerate(fh):
                line = line.strip()
                if not line or idx < start_index:
                    continue
                elif len(data) < number if number is not None else True:
                    data.append([float(i) for i in line.split(delimiter)])
        if number is not None and len(data) < number:
            raise ValueError("Not enough seed points in the file.")

        return cls(
            position=np.asarray(data),
            box_size=box_size,
            phase_label=phase_label,
        )

    def show(self) -> None:
        """
        Plot the microstructure.

        Note
        ----
        Requires matplotlib to be installed and configured.
        """
        from matplotlib import pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")
        ax.scatter(
            self.position[:, 0],
            self.position[:, 1],
            self.position[:, 2],
        )
        plt.show()



"""Test `Orientations` and related classes.

"""

from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING

from matflow.tests.utils import make_test_data_YAML_workflow

if TYPE_CHECKING:
    from matflow.param_classes.orientations import Orientations


def test_orientations_yaml_init(
    null_config, tmp_path: Path, orientations_1: Orientations
):
    wk = make_test_data_YAML_workflow("define_orientations.yaml", path=tmp_path)
    orientations = wk.tasks.define_orientations.elements[0].inputs.orientations.value
    assert orientations == orientations_1
